/*
 This file is part of the digitalSTROM ds485-core library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-core library is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation version 2
 of the License.

 digitalSTROM ds485-core library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM ds485-core library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "receiver.h"
#ifdef DAEMON_IMPL
    #include "digitalSTROM/ds485.h"
    #include "digitalSTROM/ds485-const.v1.h"
    #include "ds485-qp.v1.h"
    #include "ds485-signals.h"
#else
    #include "ds485.h"
    #include "stm32f10x.h"
#endif

#include "crc16_ccitt.h"

#ifdef Q_SPY
#include "qs_port.h"
#endif

Q_DEFINE_THIS_MODULE(receiver)

#define RECEIVER_ESCAPE_MASK    0x80

/* Active object class -----------------------------------------------------*/

typedef struct ReceiverTag {
    QActive super;
    uint16_t rxIndex;
    uint16_t rxSize;
    DS485MessageEvt* msgEvt;
    DS485LongMessageEvt* longMsgEvt;
    uint8_t header[5];
} Receiver;

static QState Receiver_initial (Receiver *me, QEvent const *e);
static QState Receiver_start_byte(Receiver *me, QEvent const *e);

static QState Receiver_receive(Receiver *me, QEvent const *e);

static QState Receiver_destination_byte(Receiver *me, QEvent const *e);
static QState Receiver_escaped_destination_byte(Receiver *me, QEvent const *e);
static QState Receiver_source_byte(Receiver *me, QEvent const *e);
static QState Receiver_escaped_source_byte(Receiver *me, QEvent const *e);
static QState Receiver_command_length_byte(Receiver *me, QEvent const *e);
static QState Receiver_escaped_command_length_byte(Receiver *me, QEvent const *e);
static QState Receiver_length_byte(Receiver *me, QEvent const *e);
static QState Receiver_escaped_length_byte(Receiver *me, QEvent const *e);

static QState Receiver_short_payload(Receiver *me, QEvent const *e);
static QState Receiver_short_payload_byte(Receiver *me, QEvent const *e);
static QState Receiver_escaped_short_payload_byte(Receiver *me, QEvent const *e);

static QState Receiver_long_payload(Receiver *me, QEvent const *e);
static QState Receiver_long_payload_byte(Receiver *me, QEvent const *e);
static QState Receiver_escaped_long_payload_byte(Receiver *me, QEvent const *e);

static void Receiver_PostShortMessage(Receiver* me);
static void Receiver_PostLongMessage(Receiver* me);
static bool Receiver_InitTokenMessage(Receiver *me, uint16_t length);
static void Receiver_InitShortDataMessage(Receiver *me, uint16_t length);
static void Receiver_InitLongDataMessage(Receiver *me, uint16_t length);
static bool Receiver_CheckCrc16(Receiver *me, uint8_t *data);

/* Local objects -----------------------------------------------------------*/
static Receiver l_receiver;     /* the single instance of the Receiver active object */

/* Global-scope objects ----------------------------------------------------*/
QActive * const AO_Receiver = (QActive *)&l_receiver;      /* "opaque" AO pointer */

/*..........................................................................*/
void Receiver_ctor(void)
{
    Receiver *me = &l_receiver;
    me->header[DS485_START_OFFSET] = DS485_START_BYTE;
    QActive_ctor(&me->super, (QStateHandler)&Receiver_initial);
}

/*..........................................................................*/
QState Receiver_initial(Receiver *me, QEvent const *e)
{
    (void)e;        /* suppress the compiler warning about unused parameter */

    /* QS dictionary entries */
    QS_OBJ_DICTIONARY(&l_receiver);

    QS_FUN_DICTIONARY(&Receiver_start_byte);

    QS_FUN_DICTIONARY(&Receiver_receive);

    QS_FUN_DICTIONARY(&Receiver_destination_byte);
    QS_FUN_DICTIONARY(&Receiver_escaped_destination_byte);
    QS_FUN_DICTIONARY(&Receiver_source_byte);
    QS_FUN_DICTIONARY(&Receiver_escaped_source_byte);
    QS_FUN_DICTIONARY(&Receiver_command_length_byte);
    QS_FUN_DICTIONARY(&Receiver_escaped_command_length_byte);
    QS_FUN_DICTIONARY(&Receiver_length_byte);
    QS_FUN_DICTIONARY(&Receiver_escaped_length_byte);

    QS_FUN_DICTIONARY(&Receiver_short_payload);
    QS_FUN_DICTIONARY(&Receiver_short_payload_byte);
    QS_FUN_DICTIONARY(&Receiver_escaped_short_payload_byte);

    QS_FUN_DICTIONARY(&Receiver_long_payload);
    QS_FUN_DICTIONARY(&Receiver_long_payload_byte);
    QS_FUN_DICTIONARY(&Receiver_escaped_long_payload_byte);

    QS_SIG_DICTIONARY(RX_START_BYTE_SIG, 0);
    QS_SIG_DICTIONARY(RX_ESCAPE_BYTE_SIG, 0);
    QS_SIG_DICTIONARY(RX_OTHER_BYTE_SIG, 0);
    QS_SIG_DICTIONARY(RX_RESERVED_BYTE_SIG, 0);

#if 0
    QS_FILTER_SM_OBJ(me);
    QS_FILTER_AO_OBJ(me);

    QS_FILTER_ON(QS_OBJ_DICTIONARY);
#endif

    return Q_TRAN(&Receiver_start_byte);
}

/*..........................................................................*/
static QState Receiver_start_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case RX_START_BYTE_SIG:
        return Q_TRAN(&Receiver_receive);
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState Receiver_receive(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_INIT_SIG:
        return Q_TRAN(&Receiver_destination_byte);
    case RX_RESERVED_BYTE_SIG:
        return Q_TRAN(&Receiver_start_byte);
    case RX_START_BYTE_SIG:
        return Q_TRAN(&Receiver_receive);
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
QState Receiver_destination_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case RX_OTHER_BYTE_SIG:
        /* save destination byte */
        me->header[DS485_DEST_OFFSET] = ((OtherByteEvt const *)e)->byte;
        return Q_TRAN(&Receiver_source_byte);
    case RX_ESCAPE_BYTE_SIG:
        return Q_TRAN(&Receiver_escaped_destination_byte);
    }
    return Q_SUPER(&Receiver_receive);
}

/*..........................................................................*/
QState Receiver_escaped_destination_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case RX_OTHER_BYTE_SIG:
        /* save destination byte */
        me->header[DS485_DEST_OFFSET] = ((OtherByteEvt const *)e)->byte | RECEIVER_ESCAPE_MASK;
        return Q_TRAN(&Receiver_source_byte);
    case RX_ESCAPE_BYTE_SIG:
        /* invalid escape byte */
        return Q_TRAN(&Receiver_start_byte);
    }
    return Q_SUPER(&Receiver_receive);
}

/*..........................................................................*/
QState Receiver_source_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case RX_OTHER_BYTE_SIG:
        /* save source byte */
        me->header[DS485_SRC_OFFSET] = ((OtherByteEvt const *)e)->byte;
        return Q_TRAN(&Receiver_command_length_byte);
    case RX_ESCAPE_BYTE_SIG:
        return Q_TRAN(&Receiver_escaped_source_byte);
    }
    return Q_SUPER(&Receiver_receive);
}

/*..........................................................................*/
QState Receiver_escaped_source_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case RX_OTHER_BYTE_SIG:
        /* save source byte */
        me->header[DS485_SRC_OFFSET] = ((OtherByteEvt const *)e)->byte | RECEIVER_ESCAPE_MASK;
        return Q_TRAN(&Receiver_command_length_byte);
    case RX_ESCAPE_BYTE_SIG:
        /* invalid escape byte */
        return Q_TRAN(&Receiver_start_byte);
    }
    return Q_SUPER(&Receiver_receive);
}

/*..........................................................................*/
QState Receiver_command_length_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        if ((me->header[DS485_DEST_OFFSET] & DS485_TYPEFLAG_MASK) == 0) {
            Receiver_InitTokenMessage(me, 3);
            Receiver_PostShortMessage(me);
            return Q_TRAN(&Receiver_start_byte);
        }
        break;
    case RX_OTHER_BYTE_SIG: {
        uint16_t length;
        /* save C/L byte */
        me->header[DS485_CL_OFFSET] = ((OtherByteEvt const *)e)->byte;
        /* get length */
        length = me->header[DS485_CL_OFFSET] & DS485_LENGTH_MASK;
        /* initialize short message */
        Receiver_InitShortDataMessage(me, length);
        if (length == 0) {
            Receiver_PostShortMessage(me);
        }
        else {
            return Q_TRAN(&Receiver_short_payload_byte);
        }
        /* invalid command */
        return Q_TRAN(&Receiver_start_byte);
    }
    case RX_ESCAPE_BYTE_SIG:
        return Q_TRAN(&Receiver_escaped_command_length_byte);
    }
    return Q_SUPER(&Receiver_receive);
}

/*..........................................................................*/
QState Receiver_escaped_command_length_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case RX_OTHER_BYTE_SIG: {
        uint16_t length;
        /* save C/L byte */
        me->header[DS485_CL_OFFSET] = ((OtherByteEvt const *)e)->byte | RECEIVER_ESCAPE_MASK;
        /* is short message? */
        length = me->header[DS485_CL_OFFSET] & DS485_LENGTH_MASK;
        /* initialize short message */
        Receiver_InitShortDataMessage(me, length);
        if (length == 0) {
            Receiver_PostShortMessage(me);
        }
        else {
            return Q_TRAN(&Receiver_short_payload_byte);
        }
        /* invalid command */
        return Q_TRAN(&Receiver_start_byte);
    }
    case RX_ESCAPE_BYTE_SIG:
        /* invalid escape byte */
        return Q_TRAN(&Receiver_start_byte);
    }
    return Q_SUPER(&Receiver_receive);
}

/*..........................................................................*/
QState Receiver_length_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case RX_OTHER_BYTE_SIG: {
        uint16_t length;
        /* save length byte */
        me->header[DS485_CL_OFFSET] = ((OtherByteEvt const *)e)->byte;
        length = me->header[DS485_CL_OFFSET];
        if (length >= DS485_MIN_DATA_PAYLOAD_SIZE) {
            if (length <= DS485_SHORT_PAYLOAD_SIZE) {
                /* short message */
                Receiver_InitShortDataMessage(me, length);
                return Q_TRAN(&Receiver_short_payload_byte);
            } else if (length <= DS485_LONG_PAYLOAD_SIZE) {
                /* long message */
                Receiver_InitLongDataMessage(me, length);
                return Q_TRAN(&Receiver_long_payload_byte);
            }
        }
        /* message too short or too long */
        return Q_TRAN(&Receiver_start_byte);
    }
    case RX_ESCAPE_BYTE_SIG:
        return Q_TRAN(&Receiver_escaped_length_byte);

    }
    return Q_SUPER(&Receiver_receive);
}

/*..........................................................................*/
QState Receiver_escaped_length_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case RX_OTHER_BYTE_SIG: {
        uint16_t length;
        /* save length byte */
        me->header[DS485_CL_OFFSET] = ((OtherByteEvt const *)e)->byte | RECEIVER_ESCAPE_MASK;
        length = me->header[DS485_CL_OFFSET];
        if (length >= DS485_MIN_DATA_PAYLOAD_SIZE) {
            if (length <= DS485_SHORT_PAYLOAD_SIZE) {
                /* short message */
                Receiver_InitShortDataMessage(me, length);
                return Q_TRAN(&Receiver_short_payload_byte);
            } else if (length <= DS485_LONG_PAYLOAD_SIZE) {
                /* long message */
                Receiver_InitLongDataMessage(me, length);
                return Q_TRAN(&Receiver_long_payload_byte);
            }
        }
        /* message too short or too long */
        return Q_TRAN(&Receiver_start_byte);
    }
    case RX_ESCAPE_BYTE_SIG:
        /* invalid escape byte */
        return Q_TRAN(&Receiver_start_byte);
    }
    return Q_SUPER(&Receiver_receive);
}

/*..........................................................................*/
QState Receiver_short_payload(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_EXIT_SIG:
        if (me->msgEvt != 0) {
            /* free allocated short message event */
            QF_gc((QEvent const *)me->msgEvt);
            me->msgEvt = 0;
        }
        return Q_HANDLED();
    }
    return Q_SUPER(&Receiver_receive);
}

/*..........................................................................*/
QState Receiver_short_payload_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case RX_OTHER_BYTE_SIG: {
        me->msgEvt->message[me->rxIndex] = ((OtherByteEvt const *)e)->byte;
        if (++me->rxIndex == me->rxSize) {
            if (Receiver_CheckCrc16(me, me->msgEvt->message)) {
                Receiver_PostShortMessage(me);
            }
            return Q_TRAN(&Receiver_start_byte);
        }
        return Q_HANDLED();
    }
    case RX_ESCAPE_BYTE_SIG:
        return Q_TRAN(&Receiver_escaped_short_payload_byte);

    }
    return Q_SUPER(&Receiver_short_payload);
}

/*..........................................................................*/
QState Receiver_escaped_short_payload_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case RX_OTHER_BYTE_SIG: {
        me->msgEvt->message[me->rxIndex] = ((OtherByteEvt const *)e)->byte | RECEIVER_ESCAPE_MASK;
        if (++me->rxIndex == me->rxSize) {
            if (Receiver_CheckCrc16(me, me->msgEvt->message)) {
                Receiver_PostShortMessage(me);
            }
            return Q_TRAN(&Receiver_start_byte);
        }
        return Q_TRAN(&Receiver_short_payload_byte);
    }
    case RX_ESCAPE_BYTE_SIG:
        /* invalid escape byte */
        return Q_TRAN(&Receiver_start_byte);

    }
    return Q_SUPER(&Receiver_short_payload);
}

/*..........................................................................*/
QState Receiver_long_payload(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_EXIT_SIG:
        if (me->longMsgEvt != 0) {
            /* free allocated short message event */
            QF_gc((QEvent const *)me->longMsgEvt);
            me->longMsgEvt = 0;
        }
        return Q_HANDLED();
    }
    return Q_SUPER(&Receiver_receive);
}

/*..........................................................................*/
QState Receiver_long_payload_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case RX_OTHER_BYTE_SIG: {
        me->longMsgEvt->message[me->rxIndex] = ((OtherByteEvt const *)e)->byte;
        if (++me->rxIndex == me->rxSize) {
            if (Receiver_CheckCrc16(me, me->longMsgEvt->message)) {
                Receiver_PostLongMessage(me);
            }
            return Q_TRAN(&Receiver_start_byte);
        }
        return Q_HANDLED();
    }
    case RX_ESCAPE_BYTE_SIG:
        return Q_TRAN(&Receiver_escaped_long_payload_byte);

    }
    return Q_SUPER(&Receiver_long_payload);
}

/*..........................................................................*/
QState Receiver_escaped_long_payload_byte(Receiver *me, QEvent const *e)
{
    switch (e->sig) {
    case RX_OTHER_BYTE_SIG: {
        me->longMsgEvt->message[me->rxIndex] = ((OtherByteEvt const *)e)->byte | RECEIVER_ESCAPE_MASK;
        if (++me->rxIndex == me->rxSize) {
            if (Receiver_CheckCrc16(me, me->longMsgEvt->message)) {
                Receiver_PostLongMessage(me);
            }
            return Q_TRAN(&Receiver_start_byte);
        }
        return Q_TRAN(&Receiver_long_payload_byte);
    }
    case RX_ESCAPE_BYTE_SIG:
        /* invalid escape byte */
        return Q_TRAN(&Receiver_start_byte);

    }
    return Q_SUPER(&Receiver_long_payload);
}

/*..........................................................................*/
static void Receiver_PostShortMessage(Receiver* me)
{
    /* update signal */
    if (me->msgEvt->message[DS485_DEST_OFFSET] & DS485_TYPEFLAG_MASK)
        me->msgEvt->super.sig = DS485_GetCommand(me->msgEvt) +  DS485_RX_SOLICIT_SUCCESSOR_REQUEST_LONG_SIG;
    else
        me->msgEvt->super.sig = DS485_RX_TOKEN_SIG;

    //printf("RX-Post: %d-%d\n", me->msgEvt->super.sig, me->msgEvt->size);

    /* send message event */
    QActive_postFIFO(AO_DS485, (QEvent*)me->msgEvt);
    /* clear message event pointer */
    me->msgEvt = 0;
}

/*..........................................................................*/
static void Receiver_PostLongMessage(Receiver* me)
{
    /* update signal */
    if (me->msgEvt->message[DS485_DEST_OFFSET] & DS485_TYPEFLAG_MASK)
        me->msgEvt->super.sig = DS485_GetCommand(me->msgEvt) +  DS485_RX_SOLICIT_SUCCESSOR_REQUEST_LONG_SIG;
    else
        me->msgEvt->super.sig = DS485_RX_TOKEN_SIG;

    //printf("RX-Post: %d-%d\n", me->msgEvt->super.sig, me->msgEvt->size);

    /* send message event */
    QActive_postFIFO(AO_DS485, (QEvent*)me->longMsgEvt);
    /* clear message event pointer */
    me->longMsgEvt = 0;
}

/*..........................................................................*/
static bool Receiver_InitTokenMessage(Receiver *me, uint16_t length)
{
    if (length == 3) {
        /* allocate message event */
        me->msgEvt = (DS485MessageEvt*) Q_NEW(DS485ControlMessageEvt, DS485_RX_TOKEN_SIG);
        /* set size */
        me->rxSize = 3;
        me->msgEvt->size = 3;
        /* copy header */
        memcpy(me->msgEvt->message, me->header, 3);
        /* OK */
        return TRUE;
    }
    return FALSE;
}

/*..........................................................................*/
static void Receiver_InitShortDataMessage(Receiver *me, uint16_t length)
{
    /* allocate message event */
    me->msgEvt = (DS485MessageEvt*) Q_NEW(DS485ControlMessageEvt, DS485_RX_REQUEST_SIG);
    /* set size */
    me->rxSize = (length + 6); /* 3 header + C/L + 2 CRC */
    me->msgEvt->size = (length + 6);
    /* copy header */
    memcpy(me->msgEvt->message, me->header, DS485_DATA_OFFSET);
    /* set receive index */
    me->rxIndex = DS485_DATA_OFFSET;
}

/*..........................................................................*/
static void Receiver_InitLongDataMessage(Receiver *me, uint16_t length)
{
    /* allocate message event */
    me->longMsgEvt = (DS485LongMessageEvt*) Q_NEW(DS485DataMessageEvt, DS485_RX_REQUEST_SIG);
    /* set size */
    me->rxSize = (length + 6); /* 3 header + C/L + 2 CRC */
    me->longMsgEvt->size = (length + 6);
    /* copy header */
    memcpy(me->longMsgEvt->message, me->header, DS485_DATA_OFFSET);
    /* set receive index */
    me->rxIndex = DS485_DATA_OFFSET;
}

/*..........................................................................*/
static bool Receiver_CheckCrc16(Receiver *me, uint8_t *data)
{
    /* check CRC-16 */
    uint16_t size = me->rxSize - 2;
    uint16_t calc_crc16 = crc16_ccitt_calc_data(0, data, size);
    uint16_t msg_crc16 = data[size+1];
    msg_crc16 <<= 8;
    msg_crc16 |= data[size];

    //printf("CRC16: is %04X - should %04X\n", msg_crc16, calc_crc16);

    if (calc_crc16 == msg_crc16) {
        return TRUE;
    }
    return FALSE;
}
