/*
 This file is part of the digitalSTROM ds485-core library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-core library is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation version 2
 of the License.

 digitalSTROM ds485-core library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM ds485-core library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/**
 * @file
 * Helper functions for tty management
 */
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include    "tty.h"
#include    "logger.h"

/**
 *    Put a dS485 line discipline in a transparent mode.
 */
static int tty_configure(struct termios *tios)
{
#if 0
    tios->c_cflag = CS8 | CREAD | B115200 | CLOCAL | HUPCL;

    /* Ignore break condition and parity errors */
    tios->c_iflag = IGNBRK | IGNPAR;
    tios->c_oflag = 0;
    tios->c_lflag = 0; /* set input mode (non-canonical, no echo,..) */
    tios->c_cc[VMIN] = 0; /* num of chars to wait for, before delivery */
    tios->c_cc[VTIME] = 0; /* timeout before delivery */

#else
    tios->c_cflag &= ~(CSIZE | PARENB | /*CRTSCTS |*/CSTOPB | IXON | IXOFF);
    tios->c_cflag |= CS8 | CREAD | CLOCAL | B115200;

    tios->c_iflag &= ~(IXON | IXOFF | IXANY | INLCR | ICRNL | IGNBRK | BRKINT | PARMRK | ISTRIP | IGNCR);
    tios->c_iflag |= INPCK;

    tios->c_oflag &= ~(OPOST);

    tios->c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);

    tios->c_cc[VMIN] = 0; /* num of chars to wait for, before delivery */
    tios->c_cc[VTIME] = 0; /* timeout before delivery */
#endif
    return (0);
}

/**
 *    Set up the serial device to be the dS485 interface.
 */
static void start_tty( Tty_t *tty, Ds485LibInst_t *dev )
{
    int cloexec;
    struct termios tios;

    /* Set new flags -> blocking mode while we configure */
    if (fcntl(tty->devfd, F_SETFL, tty->fdflags & ~O_NONBLOCK) == -1) {
        ds485_system_msg(dev, LOG_WARNING, "Couldn't set device fd flags: %m\n");
    }

    /* close-on-exec flag */
    if ((cloexec = fcntl(tty->devfd, F_GETFD)) == -1) {
        ds485_system_msg(dev, LOG_ERR, "Couldn't get device fd close-on-exec flag: %m\n");
    }
    cloexec |= FD_CLOEXEC;
    if (fcntl(tty->devfd, F_SETFD, cloexec) == -1) {
        ds485_system_msg(dev, LOG_ERR, "Couldn't set device fd close-on-exec flag: %m\n");
    }

    /* Get TTY configuration */
    if (tcgetattr(tty->devfd, &tios) != 0) {
        ds485_system_msg(dev, LOG_ERR, "tcgetattr: %m\n");
    }

    /* Save the original values */
    memcpy(&tty->termsave, &tios, sizeof(struct termios));
    tty->termvalid = 1;

    /* Setup ds485 communication parameters */
    tty_configure(&tios);
    cfsetispeed(&tios, B115200);
    cfsetospeed(&tios, B115200);

    /* Configure terminal */
    if (tcsetattr(tty->devfd, TCSANOW, &tios) < 0) {
        ds485_system_msg(dev, LOG_ERR, "tcsetattr: %m\n");
    }
    tcflush(tty->devfd, TCIFLUSH);

    /* (Re-) Enable non-blocking mode */
    if (fcntl(tty->devfd, F_SETFL, tty->fdflags | O_NONBLOCK) == -1) {
        ds485_system_msg(dev, LOG_ERR, "Couldn't set device to non-blocking mode: %m\n");
    }
}

/**
 *    Open the desired tty, save its state.
 */
int open_tty(const char *devname, Tty_t *tty, void *dev )
{
    int ttyfd;

    /* Open the serial device */
    ttyfd = open(devname, O_NONBLOCK | O_RDWR | O_NOCTTY, 0);
    tty->devfd    = ttyfd;
    if (ttyfd < 0) {
        return ttyfd;
    }

    /* Save old flags to restore them on exit */
    if ((tty->fdflags = fcntl(tty->devfd, F_GETFL)) == -1) {
        ds485_system_msg(dev, LOG_ERR, "Couldn't get device %s flags: %s\n", devname, strerror(errno));
    }
    else {
        tty->initfdflags = tty->fdflags;
    }

    /* set terminal specific config */
    start_tty(tty,dev);

    return tty->devfd;
}

/**
 * Close tty and reset termios struct
 */
int close_tty( Tty_t *tty, void *dev )
{
    if (tty->devfd != -1) {
        tty->fdflags &= ~O_NONBLOCK;
        if (fcntl(tty->devfd, F_SETFL, tty->fdflags) != 0) {
            ds485_system_msg(dev, LOG_ERR, "fcntl: fdflags: %m\n");
        }
        if (tty->termvalid)
            if (tcsetattr(tty->devfd, TCSANOW, &tty->termsave) != 0) {
                ds485_system_msg(dev, LOG_ERR, "tcsetattr: %m\n");
            }
        if (fcntl(tty->devfd, F_SETFL, tty->initfdflags) != 0) {
            ds485_system_msg(dev, LOG_ERR, "fcntl: initfdflags: %m\n");
        }
        tcflush(tty->devfd, TCIOFLUSH);
        close(tty->devfd);
        tty->devfd = -1;
    }
    return 0;
}
